# coding: utf-8

"""
Other conversions
-----------------
.. autosummary::
    :toctree: generated/

    mgc2b

"""

from __future__ import division, print_function, absolute_import

import numpy as np

from pysptk.sptk import mc2b, gnorm


def mgc2b(mgc, alpha=0.35, gamma=0.0):
    """Mel-generalized cepstrum to MGLSA filter coefficients

    Parameters
    ----------
    mgc : array, shape
        Mel-generalized cepstrum

    alpha : float
        All-pass constant. Default is 0.35.

    gamma : float
        Parameter of generalized log function. Default is 0.0.

    Returns
    -------
    b : array, shape(same as ``mgc``)
        MGLSA filter coefficients

    See Also
    --------
    pysptk.sptk.mlsadf
    pysptk.sptk.mglsadf
    pysptk.sptk.mc2b
    pysptk.sptk.b2mc
    pysptk.sptk.mcep
    pysptk.sptk.mgcep
    pysptk.sptk.mgcep

    """

    b = mc2b(mgc, alpha)
    if gamma == 0:
        return b

    b = gnorm(b, gamma)

    b[0] = np.log(b[0])
    b[1:] *= gamma

    return b



# coding: utf-8

from __future__ import with_statement, print_function, absolute_import

from setuptools import setup, find_packages, Extension
from distutils.version import LooseVersion

import numpy as np
import os
from glob import glob
from os.path import join


min_cython_ver = '0.21.0'
try:
    import Cython
    ver = Cython.__version__
    _CYTHON_INSTALLED = ver >= LooseVersion(min_cython_ver)
except ImportError:
    _CYTHON_INSTALLED = False

try:
    if not _CYTHON_INSTALLED:
        raise ImportError('No supported version of Cython installed.')
    from Cython.Distutils import build_ext
    cython = True
except ImportError:
    cython = False

if cython:
    ext = '.pyx'
    cmdclass = {'build_ext': build_ext}
else:
    ext = '.c'
    cmdclass = {}
    if not os.path.exists(join("pysptk", "_sptk" + ext)):
        raise RuntimeError("Cython is required to generate C codes.")

# SPTK sources
src_top = join("lib", "SPTK")
src_bin_top = join(src_top, "bin")
swipe_src = [
    join(src_bin_top, "pitch", "swipe", "swipe.c"),
    join(src_bin_top, "pitch", "swipe", "vector.c"),
]
rapt_src = [
    join(src_bin_top, "pitch", "snack", "jkGetF0.c"),
    join(src_bin_top, "pitch", "snack", "sigproc.c"),
]

sptklib_src = glob(join(src_top, "lib", "*.c"))
sptk_src = glob(join(src_bin_top, "*", "_*.c"))

# collect all sources
sptk_all_src = sptk_src + sptklib_src + swipe_src + rapt_src

# Filter ignore list
ignore_bin_list = [join(src_bin_top, "wavjoin"), join(src_bin_top, "wavsplit"),
                   join(src_bin_top, "vc")]
for ignore in ignore_bin_list:
    sptk_all_src = list(
        filter(lambda s: not s.startswith(ignore), sptk_all_src))

# define core cython module
ext_modules = [Extension(
    name="pysptk._sptk",
    sources=[join("pysptk", "_sptk" + ext)] + sptk_all_src,
    include_dirs=[np.get_include(), join(
        os.getcwd(), "lib", "SPTK", "include")],
    language="c",
    extra_compile_args=['-std=c99']
)]

setup(
    name='pysptk',
    version='0.1.5',
    description='A python wrapper for Speech Signal Processing Toolkit (SPTK)',
    author='Ryuichi Yamamoto',
    author_email='zryuichi@gmail.com',
    url='https://github.com/r9y9/pysptk',
    license='MIT',
    packages=find_packages(),
    ext_modules=ext_modules,
    cmdclass=cmdclass,
    install_requires=[
        'numpy >= 1.8.0',
        'six'
    ],
    tests_require=['nose', 'coverage'],
    extras_require={
        'docs': ['numpydoc', 'sphinx_rtd_theme', 'seaborn'],
        'test': ['nose'],
        'develop': ['cython >= ' + min_cython_ver],
    },
    classifiers=[
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Cython",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
    ],
    keywords=["SPTK"]
)



# coding: utf-8

"""
High-level interface for waveform synthesis
===========================================

Module ``pysptk.synthesis`` provides high-leve interface that wraps low-level
SPTK waveform synthesis functions (e.g. ``mlsadf``),

Synthesizer
-----------
.. autoclass::  Synthesizer
    :members:

SynthesisFilters
----------------

LMADF
^^^^^
.. autoclass::  LMADF
    :members:

MLSADF
^^^^^^
.. autoclass::  MLSADF
    :members:

MGLSADF
^^^^^^^
.. autoclass::  MGLSADF
    :members:

AllPoleDF
^^^^^^^^^
.. autoclass::  AllPoleDF
    :members:

AllPoleLatticeDF
^^^^^^^^^^^^^^^^
.. autoclass::  AllPoleLatticeDF
    :members:

Synthesis filter interface
--------------------------

.. autoclass:: SynthesisFilter
    :members:

"""
from __future__ import division, print_function, absolute_import

import numpy as np
import six
from abc import ABCMeta, abstractmethod

import pysptk

from pysptk.util import assert_pade, assert_stage


class SynthesisFilter(object):
    """Synthesis filter interface

    All synthesis filters must implement this interface
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def filt(self, x, coef):
        """Filter one sample

        Parameters
        ----------
        x : float
            A input sample

        coef : array
            Filter coefficients

        Returns
        -------
        y : float
            A filtered sample

        """
        pass


class Synthesizer(object):
    """Speech waveform synthesizer


    Attributes
    ----------
    filt : SynthesisFilter
        A speech synthesis filter

    hopsize : int
        Hop size

    """

    def __init__(self, filt, hopsize):
        """Initialization

        Raises
        ------
        TypeError
            if ``filt`` is not a instance of ``SynthesisFilter``

        """
        if not isinstance(filt, SynthesisFilter):
            raise TypeError("filt must be an instance of SynthesisFilter")

        self.filt = filt
        self.hopsize = hopsize

    def synthesis_one_frame(self, source, prev_b, curr_b):
        """Synthesize one frame waveform

        Parameters
        ----------
        source : array
            Source excitation

        prev_b : array
            Filter coefficients of previous frame

        curr_b : array
            Filter coefficients of current frame

        Returns
        -------
        y : array
            Synthesized waveform

        """
        assert len(prev_b) == len(curr_b)
        slope = (curr_b - prev_b) / len(source)
        interpolated_coef = prev_b.copy()

        y = np.empty_like(source)

        for i in six.moves.range(len(source)):
            scaled_source = source[i] * np.exp(interpolated_coef[0])
            y[i] = self.filt.filt(scaled_source, interpolated_coef)
            interpolated_coef += slope

        return y

    def synthesis(self, source, b):
        """Synthesize a waveform given a source excitation and sequence of
        filter coefficients (e.g. cepstrum).

        Parameters
        ----------
        source : array
            Source excitation

        b : array
            Filter coefficients

        Returns
        -------
        y : array, shape (same as ``source``)
            Synthesized waveform

        """

        y = np.zeros_like(source)

        b_prev = b[0, :]
        b_curr = b_prev
        for i in six.moves.range(b.shape[0]):
            if i > 0:
                b_prev = b[i - 1, :]
            b_curr = b[i, :]

            s, e = i * self.hopsize, (i + 1) * self.hopsize
            if e > len(source):
                break

            y[s:e] = self.synthesis_one_frame(source[s:e], b_prev, b_curr)

        return y


class LMADF(SynthesisFilter):
    """LMA digital filter that wraps ``lmadf``

    Attributes
    ----------
    pd : int
        Order of pade approximation. Default is 4.

    delay : array
        Delay

    """

    def __init__(self, order=25, pd=4):
        """Initialization

        Raises
        ======
        ValueError
            if invalid order of pade approximation is specified

        """
        self.order = order

        assert_pade(pd)

        self.pd = pd
        self.delay = pysptk.lmadf_delay(order, pd)

    def filt(self, x, coef):
        """Filter one sample using using ``lmadf``

        Parameters
        ----------
        x : float
            A input sample

        coef: array
            LMA filter coefficients (i.e. Cepstrum)

        Returns
        -------
        y : float
            A filtered sample

        See Also
        --------
        pysptk.sptk.lmadf

        """

        return pysptk.lmadf(x, coef, self.pd, self.delay)


class MLSADF(SynthesisFilter):
    """MLSA digital filter that wraps ``mlsadf``

    Attributes
    ----------
    alpha : float
        All-pass constant

    pd : int
        Order of pade approximation. Default is 4.

    delay : array
        Delay

    """

    def __init__(self, order=25, alpha=0.35, pd=4):
        """Initialization

        Raises
        ======
        ValueError
            if invalid order of pade approximation is specified

        """

        self.order = order

        assert_pade(pd)

        self.alpha = alpha
        self.pd = pd
        self.delay = pysptk.mlsadf_delay(order, pd)

    def filt(self, x, coef):
        """Filter one sample using ``mlsadf``

        Parameters
        ----------
        x : float
            A input sample

        coef: array
            MLSA filter coefficients

        Returns
        -------
        y : float
            A filtered sample

        See Also
        --------
        pysptk.sptk.mlsadf
        pysptk.sptk.mc2b

        """

        return pysptk.mlsadf(x, coef, self.alpha, self.pd, self.delay)


class MGLSADF(SynthesisFilter):
    """MGLSA digital filter that wraps ``mglsadf``

    Attributes
    ----------
    alpha : float
        All-pass constant

    stage : int
        -1/gamma

    delay : array
        Delay

    """

    def __init__(self, order=25, alpha=0.35, stage=1):
        """Initialization

        Raises
        ------
        ValueError
            if invalid number of stage is specified

        """
        self.order = order

        assert_stage(stage)

        self.alpha = alpha
        self.stage = stage
        self.delay = pysptk.mglsadf_delay(order, stage)

    def filt(self, x, coef):
        """Filter one sample using ``mglsadf``

        Parameters
        ----------
        x : float
            A input sample

        coef: array
            MGLSA filter coefficients

        Returns
        -------
        y : float
            A filtered sample

        See Also
        --------
        pysptk.sptk.mglsadf

        """

        return pysptk.mglsadf(x, coef, self.alpha, self.stage, self.delay)


class AllPoleDF(SynthesisFilter):
    """All-pole digital filter that wraps ``poledf``

    Attributes
    ----------
    delay : array
        Delay

    """

    def __init__(self, order=25):
        """Initialization
        """

        self.order = order
        self.delay = pysptk.poledf_delay(order)

    def filt(self, x, coef):
        """Filter one sample using using ``poledf``

        Parameters
        ----------
        x : float
            A input sample

        coef: array
            LPC (with loggain)

        Returns
        -------
        y : float
            A filtered sample

        See Also
        --------
        pysptk.sptk.poledf

        """

        return pysptk.poledf(x, coef, self.delay)


class AllPoleLatticeDF(SynthesisFilter):
    """All-pole lttice digital filter that wraps ``ltcdf``

    Attributes
    ----------
    delay : array
        Delay

    """

    def __init__(self, order=25):
        """Initialization
        """

        self.order = order
        self.delay = pysptk.ltcdf_delay(order)

    def filt(self, x, coef):
        """Filter one sample using using ``ltcdf``

        Parameters
        ----------
        x : float
            A input sample

        coef: array
            PARCOR coefficients (with loggain)

        Returns
        -------
        y : float
            A filtered sample

        See Also
        --------
        pysptk.sptk.ltcdf

        """

        return pysptk.ltcdf(x, coef, self.delay)


class LSPDF(SynthesisFilter):
    """All-pole digital filter that wraps ``lspdf``

    Attributes
    ----------
    delay : array
        Delay

    """

    def __init__(self, order=25):
        """Initialization
        """

        self.order = order
        self.delay = pysptk.lspdf_delay(order)

    def filt(self, x, coef):
        """Filter one sample using using ``lspdf``

        Parameters
        ----------
        x : float
            A input sample

        coef: array
            LSP (with loggain)

        Returns
        -------
        y : float
            A filtered sample

        See Also
        --------
        pysptk.sptk.lspdf

        """

        return pysptk.lspdf(x, coef, self.delay)



# coding: utf-8

"""
Utilities
=========

Audio files
-----------

.. autosummary::
    :toctree: generated/

    example_audio_file

"""

from __future__ import division, print_function, absolute_import

import pkg_resources

# 16kHz, 16bit example audio from cmu_us_awb_arctic
# see COPYING for the license of the audio file.
EXAMPLE_AUDIO = 'example_audio_data/arctic_a0007.wav'


def assert_gamma(gamma):
    if not (-1 <= gamma <= 0.0):
        raise ValueError("unsupported gamma: must be -1 <= gamma <= 0")


def assert_pade(pade):
    if pade != 4 and pade != 5:
        raise ValueError("4 or 5 pade approximation is supported")


def assert_stage(stage):
    if stage < 1:
        raise ValueError("stage >= 1 (-1 <= gamma < 0)")


def ispow2(num):
    return ((num & (num - 1)) == 0) and num != 0


def assert_fftlen(fftlen):
    if not ispow2(fftlen):
        raise ValueError("fftlen must be power of 2")


def example_audio_file():
    """Get the path to an included audio example file.

    Examples
    --------
    >>> from scipy.io import wavfile
    >>> fs, x = wavfile.read(pysptk.util.example_audio_file())

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(x, label="cmu_us_awb_arctic arctic_a0007.wav")
    >>> plt.xlim(0, len(x))
    >>> plt.legend()

    """

    return pkg_resources.resource_filename(__name__, EXAMPLE_AUDIO)
